/***
chatClient: Cliente que se conecta a la
sala de chat.

Gabriel Ahumada 2821039-6
Constanza Valds 2821143-0

Basado en la solucin de Daniel Doris.
***/

#include <sys/types.h> 
#include <pthread.h> 
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <netdb.h> 
#include <stdio.h> 
#include <string.h> 
#include <pwd.h> 
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/time.h>

#define BUFFSIZE 1024 // Tamao del buffer
#define NOMSIZE 16 // Tamao mximo del nombre de usuario
#define HOSTSIZE 64 // Tamao del string hostname

pid_t pid;

/* Funcin que realiza el hilo para recibir los mensajes 
del socket y desplegarlo por pantalla */
void* hiloLectura (void* arg) {
	int *socket; // Variable local para el socket
    char buf[BUFFSIZE]; // Buffer de lectura
    int n; // Nmero de bytes recibidos
	
	// Asignamos el argumento de entrada
    socket = arg;
    
	// Lo que se recibe del socket se muestra por pantalla
    while ((n = recv(*socket, buf, sizeof(buf), 0)) > 0){
        write(1, buf, n);
    }
    // Terminada la conexin se elimina el proceso hijo
    kill(pid, SIGKILL);
    // Y se elimina el socket
    close(*socket);
    exit(0);
}
    
    
/* Funcin principal que crea un proceso hijo
para mostrar mensajes en otra pantalla e invoca la hebra de lectura */
int main(int argc, char * argv[]) {
	int s; // Socket de conexin con el servidor
    int n, len, port; // Variable axiliar y puerto
    char buf[BUFFSIZE]; // Buffer de escritura
    char hostname[HOSTSIZE]; // Buffer del hostname
    struct hostent *hp; // Estructura para el usuario
    struct sockaddr_in name; // Estructura para el socket
    int pfd[2]; // Pipe para la ventana
    pthread_t myTid; // Tid de la hebra de lectura
	struct passwd* myPasswd; // Estructura para el nombre del usuario
	uid_t myUid; // ID del usuario
	char adios[]="bye"; // Mensaje de salida

	// Se analizan los argumentos de entrada
    switch(argc) {
		case 3:
			// Si se ingresan dos argumentos, se asigna hostname y puerto 
			strcpy(hostname, argv[1]);
			port = atoi(argv[2]);
			if(port > 0) break;
		
		default:
			// Se muestra el uso esperado
			printf("Usage: %s [nombre_servidor] [puerto]\n", argv[0]);
			exit(1);
    }
	
	// Se obtiene la uid
	myUid = getuid();
	// Se obtiene el nombre del usuario desde la estructura passwd
	myPasswd = getpwuid(myUid);
	if (myPasswd == NULL) {
		perror("getpwuid()");
		exit(1);
	}	
	// Se guarda el nombre de usuario en buf
	n = strlen(myPasswd->pw_name);
	memcpy(buf, myPasswd->pw_name, n+1);
	buf[n] = '\0';

	// Se informa por pantalla el intento de conexin
	printf("Conectando a: %s, en puerto: %d. Usuario: %s.\n", hostname, port, buf);
    
    // Se obtienen los datos del servidor
    hp = gethostbyname(hostname);
	if(hp == NULL) {
		herror("gethostbyname()"); 
		exit(1);
	}
    // Se crea el socket
    s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		perror("socket()");
		exit(1);
	}
    // Se crea la direccin del servidor
    name.sin_family = AF_INET;
    name.sin_port = htons(port);
    memcpy(&name.sin_addr, hp->h_addr_list[0], hp->h_length);
    len = sizeof(struct sockaddr_in);

    // Se intenta la conexin al servidor
    if (connect(s, (struct sockaddr*) &name, len)) {
		perror("connect()");
		exit(1);
	}
    
    // Se crea el pipe (tubera)
	if(pipe(pfd) < 0){
        perror("pipe");
        exit(1);
    }
    
    // Se crea el proceso hijo 
    if((pid = fork()) < 0){
        perror("fork");
        exit(1);
    }

    //Cdigo ejecutado por el hijo
    if(pid == 0){
    	// Se conecta la entrada estndar del hijo con la salida del pipe
        if (dup2(pfd[0],0) == -1) {
			perror("fallo en dup2");
			exit(1);
        }
        // Se cierra la entrada al pipe
      	close(pfd[1]);
      	fflush(stdin);       
    	
    	// Se obtiene el directorio actual
		char cwd[1024];
        if (getcwd(cwd, sizeof(cwd)) == NULL){
        	perror("getcwd");
        	exit(1);
        }
        sprintf(cwd, "%s/lector", cwd);
        // Se ejecuta el archivo binario del cdigo fuente lector.c
        execlp("xterm", "-e", cwd, NULL);
        // En caso de error
      	perror("execl");
      	_exit(127);
    }

    // Se conecta la salida estndar a la entrada del pipe
   	if (dup2(pfd[1],1) == -1) {
			perror("fallo en dup2");
			exit(1);
    }
    // Se cierra la salida del pipe
    close(pfd[0]);            

	// Se crea un nuevo hilo para leer desde el servidor
    if (pthread_create (&myTid, NULL, hiloLectura, &s)) {
        perror("pthread_create()");
        exit(1);
    }
    // No se espera su valor de retorno
    if (pthread_detach(myTid)) {
        perror("pthread_detach()");
        exit(1);
    }

    // Se enva lo escrito desde entrada estndar hacia el servidor
    n = send(s, buf, strlen(buf)+1, 0);
    while ((n = read(0, buf, sizeof(buf))) > 0) {
    	// Si ingresa la palabra "bye" se cierra la conexin
    	if(memcmp(adios, buf, 3) == 0) break;
        if (send(s, buf, n, 0) < 0) {
        	kill(pid, SIGKILL);
            perror("send");
            exit(1);
        }
    }
    
	// Destruccin del socket y trmino al proceso hijo
    close(s);
    kill(pid, SIGKILL);
    exit(0);
}
