/***
chatMonitor: Cliente especial del monitor
que entrega información acerca de las conexiones.

Gabriel Ahumada 2821039-6
Constanza Valdés 2821143-0

Basado en la solución de Daniel Doris.
***/

#include <sys/types.h> 
#include <pthread.h> 
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <netdb.h> 
#include <stdio.h> 
#include <string.h> 
#include <pwd.h> 
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/time.h>

#define BUFFSIZE 1024 // Tamaño del buffer
#define NOMSIZE 16 // Tamaño máximo del nombre de usuario
#define HOSTSIZE 64 // Tamaño del string hostname


/* Función que realiza el hilo para recibir los mensajes 
del socket y desplegarlo por pantalla */
void* hiloLectura (void* arg) {
    int *socket; // Variable local para el socket
    char buf[BUFFSIZE]; // Buffer de lectura
    int n; // Número de bytes recibidos
    
    // Asignamos el argumento de entrada
    socket = arg;
    
    // Lo que se recibe del socket se muestra por pantalla
    while ((n = recv(*socket, buf, sizeof(buf), 0)) > 0){
        write(1, buf, n);
    }
    // Y se elimina el socket
    close(*socket);
    exit(0);
}
    
    
/* Función principal que muestra mensajes por pantalla 
e invoca la hebra de lectura */
int main(int argc, char * argv[]) {
	int s; // Socket de conexión con el servidor
    int n, len, port; // Variable axiliar y puerto
    char buf[BUFFSIZE]; // Buffer de escritura
    char hostname[HOSTSIZE]; // Buffer del hostname
    struct hostent *hp; // Estructura para el usuario
    struct sockaddr_in name; //Estructura para el socket
    pthread_t myTid; // Tid de la hebra de lectura

	// Se analizan los argumentos de entrada
    switch(argc) {
		case 2:
            // Si sólo ingresa un argumento, el hostname es la máquina misma
			if (gethostname(hostname, sizeof(hostname))){
				perror("gethostname()");
				exit(1);
			}
			port = atoi(argv[1]);
			if (port > 0) break;
		default:
			// En caso de tener más argumentos, se muestra el uso esperado
			printf("Usage: %s [puerto_monitor]\n", argv[0]);
			exit(1);
    }
	
	// Se informa por pantalla
	printf("Conectando a %s en el puerto: %d.\n", hostname, port);
    // Se obtienen los datos del servidor
    hp = gethostbyname(hostname);
	if(hp == NULL) {
		herror("gethostbyname()"); 
		exit(1);
	}
    // Se crea el socket de conexión
    s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		perror("socket()");
		exit(1);
	}
    // Se crea la dirección del servidor
    name.sin_family = AF_INET;
    name.sin_port = htons(port);
    memcpy(&name.sin_addr, hp->h_addr_list[0], hp->h_length);
    len = sizeof(struct sockaddr_in);

    // Se intenta la conexión con el servidor
    if (connect(s, (struct sockaddr*) &name, len)) {
		perror("connect()");
		exit(1);
	}
	// Se crea un nuevo hilo para leer desde el servidor
    if (pthread_create (&myTid, NULL, hiloLectura, &s)) {
        perror("pthread_create()");
        exit(1);
    }
    // No se espera su valor de retorno
    if (pthread_detach(myTid)) {
        perror("pthread_detach()");
        exit(1);
    }

    // Se envía lo escrito desde entrada estándar hacia el servidor
    n = send(s, buf, strlen(buf)+1, 0);
    while ((n = read(0, buf, sizeof(buf))) > 0) {
        if (send(s, buf, n, 0) < 0) {
            perror("send");
            exit(1);
        }        
    }
    
	// Destrucción del socket
    close(s);
    exit(0);
}
