/***
chatServer: Servidor que administra las conexiones 
con los clientes y el monitor.

Gabriel Ahumada 2821039-6
Constanza Valds 2821143-0

Basado en la solucin de Daniel Doris.
***/

#include <sys/types.h> 
#include <pthread.h> 
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <stdio.h> 
#include <string.h> 
#include <unistd.h> 
#include <stdlib.h> 
#include <arpa/inet.h>
#include <errno.h>

#define MAXUSUARIOS 5 // Nmero mximo de usuarios en el servidor
#define NOMSIZE 16 // Tamao mximo del nombre de usuario
#define BUFSIZE 1024 // Tamao del buffer de mensajes
#define T_ESPERA 5000 // Tiempo en microsegundos que duerme el hilo reenviante

int sockArr[MAXUSUARIOS+1]; // Arreglo de sockets
int usrNum; // Nmero de usuarios conectados
char buffRenv[BUFSIZE*MAXUSUARIOS*5]; // Buffer de reenvo con un tope de cinco mensajes por usuario 
int n_bytes[MAXUSUARIOS*5]; //Nmero de bytes tranferidos
int buffNum; // ndice del buffer de reenvo
int mayorSock=0; //Socket de mayor nmero
pthread_mutex_t sockMutex; // Mutex para acceder al arreglo de sockets
pthread_mutex_t buffMutex; // Mutex para acceder el buffer de reenvo
pthread_mutex_t buffMutex_monitor; //Mutex para acceder al buffer del monitor


/* Esta funcin es ejecutada por los hilos que atienden a los usuarios */
void* hiloServidor (void* arg) {
    int* socket; //Variable local para el socket
    int indice; // ndice que apunta al final del nombre de usuario en el mensaje
	char mensaje[NOMSIZE+BUFSIZE+1]; // Buffer que contiene el nombre de usuario ms el mensaje
    int n; //Nmero de bytes recibidos
    socket = (int*) arg; // Asignacin de socket desde el argumento
    n_bytes[*socket]=0; //Arreglo que guarda en la posicin de socket el nmero total de bytes tranferidos

    //Bsqueda del mayor socket asignado
    if(*socket>mayorSock) mayorSock=*socket; 
    // Se obtiene el nombre de usuario
    n = recv(*socket, mensaje, sizeof(mensaje), 0); 
    // Si se recibe correctamente
    if (n > 0) { 
        //Aumenta de forma segura el nmero de bytes transferidos
        pthread_mutex_lock(&buffMutex_monitor);
        n_bytes[*socket] = n_bytes[*socket] + n;
        pthread_mutex_unlock(&buffMutex_monitor);
        
        // Se anuncia al servidor
        printf("Usuario: %s. Socket: %d\n", mensaje, *socket);
        // Se anuncia de forma segura al resto de los usuarios
        
        pthread_mutex_lock(&buffMutex);
        // Si no hay espacio en el buffer, se avisa
        if (buffNum+n+30 >=  sizeof(buffRenv))
            printf("Error: buffer lleno. Se descartara el mensaje para evitar inconsistencias.\n");
        // Si hay espacio escribimos en el buffer
        else {
            memcpy(mensaje+n, " se ha conectado al servidor.\n", 30), 
            memcpy(buffRenv+buffNum, mensaje, n+30);
            buffNum += n+30;
            buffRenv[buffNum] = '\0';
        }
        pthread_mutex_unlock(&buffMutex);
        
        // Se construye el mensaje "nombre_de_usuario: "
        mensaje[n-1] = ':';
        mensaje[n] = ' ';
        // Se guarda el ndice
        indice = n+1;
    }
    
    // Si no se recibe correctamente, la hebra es eliminada
    else {
        // Se elimina un socket de forma segura
        pthread_mutex_lock(&sockMutex);
        // Se borra de la lista y disminuye el nmero de usuarios
        close(*socket);
        *socket = 0;
        usrNum--;
        pthread_mutex_unlock(&sockMutex);
        pthread_exit(NULL);
    }
	
    // Se lee del socket, mientras existan posibles mensajes
	while ((n = recv(*socket, mensaje +indice, sizeof(mensaje), 0)) > 0) {
        //Aumenta de forma segura el nmero de bytes transferidos
        pthread_mutex_lock(&buffMutex_monitor);
        n_bytes[*socket] = n_bytes[*socket] + n;
        pthread_mutex_unlock(&buffMutex_monitor);

        // Se copia el mensaje al buffer de reenvo
        pthread_mutex_lock(&buffMutex);
        // Si no hay espacio en el buffer se avisa
        if (buffNum+n+indice >=  sizeof(buffRenv))
            printf("Error: buffer lleno. Se descartar el mensaje para evitar inconsistencias.\n");
        // Si hay espacio se escribe en el buffer
        else {
            memcpy(buffRenv+buffNum, mensaje, n+indice);
            buffNum += n+indice;
            buffRenv[buffNum] = '\0';
        }
        pthread_mutex_unlock(&buffMutex);
    }
    // Terminada la conexin la hebra muere
	mensaje[indice-2] = '\0';
	// Se anuncia al servidor
	printf("Usuario %s se ha desconectado. Socket: %d\n", mensaje, *socket);
	
	// Se anuncia al resto de los usuarios
	pthread_mutex_lock(&buffMutex);
	// Si no hay espacio en el buffer se avisa
	if (buffNum+n+30 >=  sizeof(buffRenv))
		printf("Error: buffer lleno. Se descartara el mensaje para evitar inconsistencias.\n");
	// Si hay espacio se escribe en el buffer
	else {
		memcpy(mensaje+indice-2, " se ha desconectado.\n", 21), 
		memcpy(buffRenv+buffNum, mensaje, indice+19);
		buffNum += indice+19;
		buffRenv[buffNum] = '\0';
	}
	pthread_mutex_unlock(&buffMutex);

    // Se elimina un socket de forma segura 
    pthread_mutex_lock(&sockMutex);
    // Se borra de la lista y disminuye el nmero de usuarios
    close(*socket);
    *socket = 0;
    usrNum--;
    pthread_mutex_unlock(&sockMutex);
    pthread_exit(NULL);
}


/* Esta funcin es ejecutada por los hilos 
para el reenvo de los mensajes a todos los usuarios */
void* hiloReenvio (void* arg) {
    int i; // Iterador
    
    while (1) {
        // Se duerme 5 [ms]
        usleep(T_ESPERA);
        // En forma segura se lee del buffer
        pthread_mutex_lock(&buffMutex);      
        // Si existen mensajes pendientes
        if (buffNum > 0) {
            pthread_mutex_lock(&sockMutex);
            // Se envan los mensajes a todos los sockets vlidos
            for(i = 0; i < MAXUSUARIOS+1; i++){
                if (sockArr[i] > 0){
                    send (sockArr[i], buffRenv, buffNum, 0);
                }
            }
            pthread_mutex_unlock(&sockMutex);
            // Se resetea el ndice
            buffNum = 0;
        }
        pthread_mutex_unlock(&buffMutex);
    }
}

/* Funcin que ejecuta un hebra
para atender las peticiones del monitor */
void* hiloMonitor (void* arg) {
    
int *port_monitor; // Variable local para el puerto
char mensaje[BUFSIZE+1]; // Arreglo para el mensaje
char estado[]="STATE"; // Comando STATE 
char fin[]="STOP"; // Comando STOP
char bienvenida[]="Por favor ingrese STATE o STOP:\n"; // Mensaje de bienvenida
char despedida[]="Fin de la sesion.\n"; // Mensaje de despedida
char nobody[]="No hay conexiones con chatServer.\n"; // Mensaje en el caso que no haya usuarios en el servidor
char conexiones[40]; // Mensaje para el nmero total de conexiones
char reporte[70]; // Mensaje para el reporte de bytes transferidos 
int sockMon; // Descriptor de archivo para el socket
int s_m; // Socket de monitor
struct sockaddr_in name_monitor; // Estructura propia del socket para el monitor
socklen_t len=sizeof(struct sockaddr_in); 
int n, j, i=1;

// Asignacin del puerto desde el argumento
port_monitor = arg;
    
    // Ciclo que slo termina si el programa principal lo hace
    while(1){ 
         // Se crea el socket para el monitor
        s_m = socket(AF_INET, SOCK_STREAM, 0);
        if (setsockopt(s_m, SOL_SOCKET, SO_REUSEADDR, (char*) &i, sizeof(i)) == -1) {
            perror("setsockopt()");
            exit(1);
        }

        // Se crea la direccin del servidor
        name_monitor.sin_family = AF_INET;
        name_monitor.sin_port = htons(*port_monitor);
        name_monitor.sin_addr.s_addr = htonl(INADDR_ANY);
        
        // Se liga la direccin al socket
        if (bind (s_m, (struct sockaddr*) &name_monitor, len)) {
            perror("bind");
            exit(1);
        }
        // Se obtiene el puerto finalmente asignado
        getsockname(s_m, (struct sockaddr*) &name_monitor, &len);
        printf("Escuchando a monitor en el puerto %d\n", ntohs(name_monitor.sin_port));
        
        // Se espera la conexin del monitor
        if (listen(s_m, 5)) { 
            perror("listen()");
            exit(1);
        }
        // Se acepta la conexin 
        sockMon = accept(s_m, (struct sockaddr*) &name_monitor, &len);
        if (sockMon < 0) {
            perror("accept()");
            exit(1);
        }

        printf("Recibida una conexion desde : %s\n",inet_ntoa(name_monitor.sin_addr));
        
        // Se lee del socket, mientras existan posibles mensajes
        while ((n = recv(sockMon, mensaje, sizeof(mensaje), 0)) > 0) {
            // Si escribe STATE
            if(memcmp(estado, mensaje, 5) == 0){ 
                // Y existen usuarios conectados
                if(usrNum>0){
                    pthread_mutex_lock(&buffMutex_monitor);
                    sprintf(conexiones, "El numero total de conexiones es: %d\n", usrNum);
                    send (sockMon, conexiones, sizeof(conexiones), 0);
                    // Se enva la informacin de estos
                    for(j = 0; j <= mayorSock; j++)
                        if (n_bytes[j] > 0){
                            sprintf(reporte, "El cliente con el socket %d ha tranferido un total de %d bytes\n", j, n_bytes[j]);
                            send (sockMon, reporte, sizeof(reporte), 0);
                        }
                    pthread_mutex_unlock(&buffMutex_monitor);
                } 
                // Si no, se avisa chatServer vaco
                else{
                    send(sockMon, nobody, sizeof(nobody),0);  
                }
            }   
            // Si escribe STOP, se sale del ciclo
            else if (memcmp(fin, mensaje, 4) == 0) break;
            // Si escribe otra palabra se sugiere STOP y STATE
            else {
                printf("Esperando comando correcto desde el monitor... \n");
                send(sockMon, bienvenida, sizeof(bienvenida),0);
            }
        }
        //Se elimina el socket
        printf("Conexion terminada con el monitor\n");
        send(sockMon, despedida, sizeof(despedida),0);
        close(sockMon);
        close(s_m);
    }
    pthread_exit(NULL);
}


/* Funcin principal que invoca a los hilos
y recibe  a los clientes */
int main (int argc, char * argv[]) {

	int i, sockCont, port, port_monitor; // Variable auxiliar, contador de socket y puertos
	pthread_t tid[MAXUSUARIOS+1], tidReenvio, tidMonitor; // Hebras a invocar
    int s; // Socket de clientes
    struct sockaddr_in name; // Estructura propia del socket para el monitor 
    socklen_t len; 
    
	// Se analizan los argumentos de entrada
    switch(argc) {
		case 1:
			// Si no se ingresan argumentos, el SO asigna los puertos
			port = 0;
            port_monitor = 0;
			break;
		case 2:
			// Si slo se ingresa un argumento corresponde al puerto de clientes
			port = atoi(argv[1]);
            port_monitor = 0;
			if (port > 0) break;
		case 3:
            // Si se ingresan dos argumentos cada uno corresponde al puerto de clientes y monitor
            port = atoi(argv[1]);
            port_monitor = atoi(argv[2]);
            if (port > 0 && port_monitor > 0) break; 
		default:
			// En caso de tener ms argumentos, mostramos el uso esperado
			printf("Usage: %s [puerto_chat] [puerto_monitor]\n", argv[0]);
			exit(1);
    }
	
    // Se inicializan los mutex
    pthread_mutex_init(&sockMutex, NULL);
	pthread_mutex_init(&buffMutex, NULL);
    pthread_mutex_init(&buffMutex_monitor, NULL);
    
    // Se inicializan los sockets en 0
    for (sockCont = 0; sockCont < MAXUSUARIOS+1; sockCont++){
        sockArr[sockCont] = 0;
        n_bytes[sockCont] =0;
    }
    sockCont = 0;
    
	// Se crea el socket para los clientes
    s = socket(AF_INET, SOCK_STREAM, 0);
	i = 1;
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char*) &i, sizeof(i)) == -1) {
		perror("setsockopt()");
		exit(1);
	}
    // Se crea la direccin del servidor
    name.sin_family = AF_INET;
    name.sin_port = htons(port);
    name.sin_addr.s_addr = htonl(INADDR_ANY);
    len = sizeof(struct sockaddr_in);

    // Se liga la direccin al socket
   if (bind (s, (struct sockaddr*) &name, len)) {
		perror("bind");
		exit(1);
	}

    // Se obtiene el puerto a utilizar y se muestra
    getsockname(s, (struct sockaddr*) &name, &len);
    printf("chatServer online\n");
    printf("Escuchando a clientes en el puerto %d\n", ntohs(name.sin_port));
   
    // Se crea el hilo para las funciones del monitor
    if (pthread_create (&tidMonitor, NULL, hiloMonitor, &port_monitor)) {
        perror("pthread_create()");
        exit(1);
    }
    // No se espera su valor de retorno
    if (pthread_detach(tidMonitor)) {
        perror("pthread_detach()");
        exit(1);
    }
    
    // Se inicia el hilo de reenvo
    if (pthread_create (&tidReenvio, NULL, hiloReenvio, NULL)) {
        perror("pthread_create()");
        exit(1);
    }
    // No se espera su valor de retorno
    if (pthread_detach(tidReenvio)) {
        perror("pthread_detach()");
        exit(1);
    }
    
   
    // Espera de clientes
	while (1) {
        
		// Se espera por una conexin
		if (listen(s, 5)) {
			perror("listen()");
			exit(1);
		}

		// Se acepta la conexin
		sockArr[sockCont] = accept(s, (struct sockaddr*) &name, &len);
		if (sockArr[sockCont] < 0) {
			perror("accept()");
			exit(1);
		}

        pthread_mutex_lock(&sockMutex);
        // Si hay ms cupos
        if (usrNum < MAXUSUARIOS) {
            
            printf("Recibida una conexion desde : %s. ",inet_ntoa(name.sin_addr));
            printf("Cantidad de usuarios: %d.\n", usrNum+1);
            // Se crea un nuevo hilo para ese usuario
            if (pthread_create (&tid[sockCont], NULL, hiloServidor, &sockArr[sockCont])) {
                perror("pthread_create()");
                exit(1);
            }
            // No se espera su valor de retorno
            if (pthread_detach(tid[sockCont])) {
                perror("pthread_detach()");
                exit(1);
            }
            // Se considera en la lista
            usrNum++;
            // Se pasa al siguiente socket libre
            do {
                sockCont++;
                sockCont %= MAXUSUARIOS+1;
            } while (sockArr[sockCont] != 0);
        }
        
        // Si no hay cupos
        else {
            printf("Recibida una conexion desde : %s. ", inet_ntoa(name.sin_addr));
            printf("Servidor lleno (%d usuarios). Se negara el servicio.\n", usrNum+1);
            
            // Se le notifica al cliente
            send (sockArr[sockCont], "Error: servidor lleno.\n", 24, 0);
            // Y se cierra la conexin
            send (sockArr[sockCont], NULL, 0, 0);
            close (sockArr[sockCont]);
            sockArr[sockCont] = 0;
        }
        pthread_mutex_unlock(&sockMutex);
	}
}
